// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <i.h>
#include <math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_multifit_nlin.h>
static struct istruct global_in;
static struct istruct global_sim_i;
static struct istruct global_r;
static struct istruct global_A;
static struct istruct global_dos;
static struct istruct global_sigma;
static struct istruct global_error;


#define mypoints 100

static  int pt_M=mypoints;
static int pt_N=mypoints*mypoints;
static int *pt_Ti;
static int *pt_Tj;
static double *pt_Tx;
static double *pt_b;

static int dolog=FALSE;
struct pt_data {
size_t n;
double * y;
double * sigma;
};

void pt_update_dos(const gsl_vector * x)
{
int i;
for (i=0;i<global_dos.len;i++)
{

		global_dos.data[i]=fabs(gsl_vector_get(x, i));
}
}

double gen_i(struct istruct *r,struct istruct *A,double t)
{
double sum;
int i;
sum=0.0;
for (i=0;i<global_dos.len;i++)
{
	sum+=fabs(global_dos.data[i])*r->data[i]*exp(-r->data[i]*t);
//printf("%d %le %le %le %le %le %le\n",i,global_dos.data[i],r->data[i],exp(-r->data[i]*t),r->data[i],t,global_dos.data[i]*r->data[i]*exp(-r->data[i]*t));
//getchar();
}
//printf("%le\n",sum);
//getchar();
//printf("break\n");


return sum*1e-12;
}

double get_pt_error()
{
int i;
double error=0.0;
//inter_save_a(&global_sim_i,"./","test0.dat");
for (i=0;i<global_in.len;i++)
{
global_sim_i.x[i]=global_in.x[i];
global_sim_i.data[i]=gen_i(&global_r,&global_A,global_in.x[i]);
}

//getchar();
if (dolog==TRUE) inter_log_y(&global_sim_i);
//inter_save_a(&global_sim_i,"./","test1.dat");
//inter_save_a(&global_dos,"./","test2.dat");

inter_save_a(&global_sim_i,"./","pt_fit_i.dat");

for (i=0;i<global_in.len;i++)
{
global_error.data[i]=fabs(global_sim_i.data[i]-global_in.data[i]);
global_error.x[i]=global_in.x[i];

error+=fabs(global_error.data[i]);
}

return error;
}

void do_newton()
{
printf("do newton\n");


size_t i;
int ii;
int pos=0;
for (i = 0; i <global_sim_i.len; i++)
{

	/* Jacobian matrix J(i,j) = dfi / dxj, */
	/* where fi = (Yi - yi)/sigma[i],      */
	/*       Yi = A * exp(-lambda * i) + b  */
	/* and the xj are the parameters (A,lambda,b) */
	for (ii=0;ii<global_dos.len;ii++)
	{
		get_pt_error ();
//inter_save_a(&global_error,"./","temp_error1.dat");
		double y0=0.0;

		y0=global_error.data[i];
		double save=global_dos.data[ii];

		double delta=fabs(global_dos.data[ii]*0.001);
		global_dos.data[ii]+=delta;
		get_pt_error ();
//inter_save_a(&global_error,"./","temp_error2.dat");
//inter_save_a(&global_dos,"./","glob_dos.dat");
		double y1=0.0;

		y1=global_error.data[i];

		double deriv=(y1-y0)/delta;
		//printf("%d %d %le\n",i,ii,global_dos.data[ii]);
		//getchar();

		pt_Ti[pos]=i;
		pt_Tj[pos]=ii;
		pt_Tx[pos]=deriv;
		pos++;
		//gsl_matrix_set (J, i, ii, deriv/s);
		//printf("%d %d %le %le\n",i,ii,y1,y0);
		global_dos.data[ii]=save;
//getchar();
	}
//getchar();
printf("J= %d %d\n",(int)i,(int)ii);

}

inter_save_a(&global_in,"./","pt_exp_i.dat");
inter_save_a(&global_dos,"./","pt_fit_dos.dat");
inter_save_a(&global_sim_i,"./","pt_fit_i.dat");


inter_save_a(&global_sim_i,"./","ittr_sim_i.dat");
inter_save_a(&global_in,"./","ittr_in.dat");
inter_save_a(&global_dos,"./","ittr_dos.dat");
inter_save_a(&global_error,"./","ittr_error.dat");
//getchar();

get_pt_error ();

for (i=0;i<global_error.len;i++)
{
pt_b[i]=-global_error.data[i];
}

solver_dump_matrix(mypoints,mypoints*mypoints,pt_Ti,pt_Tj, pt_Tx,pt_b);
//getchar();
solver(pt_M,pt_N,pt_Ti,pt_Tj, pt_Tx,pt_b);

double update=0.0;
for (i=0;i<global_dos.len;i++)
{
	global_dos.data[i]+=pt_b[i]/(1.0+fabs(pt_b[i]/1e3));
	printf("%d %le\n",i,pt_b[i]);
	update+=fabs(pt_b[i]);
}

printf("exit expb_df %le\n",update);

}

int pt_expb_df (const gsl_vector * x, void *data, gsl_matrix * J)
{

double *sigma = global_sigma.data;
printf("pt_expb_df\n");
pt_update_dos(x);

size_t i;
int ii;

for (i = 0; i <global_sim_i.len; i++)
{
	double s = sigma[i];

	/* Jacobian matrix J(i,j) = dfi / dxj, */
	/* where fi = (Yi - yi)/sigma[i],      */
	/*       Yi = A * exp(-lambda * i) + b  */
	/* and the xj are the parameters (A,lambda,b) */
	for (ii=0;ii<global_dos.len;ii++)
	{
		get_pt_error ();
//inter_save_a(&global_error,"./","temp_error1.dat");
		double y0=0.0;

		y0=global_error.data[i];
		double save=global_dos.data[ii];

		double delta=fabs(global_dos.data[ii]*0.01);
		global_dos.data[ii]+=delta;
		get_pt_error ();
//inter_save_a(&global_error,"./","temp_error2.dat");
//inter_save_a(&global_dos,"./","glob_dos.dat");
		double y1=0.0;

		y1=global_error.data[i];

		double deriv=(y1-y0)/delta;
		//printf("%d %d %le\n",i,ii,global_dos.data[ii]);
		//getchar();
		gsl_matrix_set (J, i, ii, deriv/s);
		//printf("%d %d %le %le\n",i,ii,y1,y0);
		global_dos.data[ii]=save;
//printf("J= %d %d %le %le %le %le %le %le\n",i,ii,save,deriv,y1,y0,global_dos.data[ii],delta);
//getchar();
	}


printf("J = %d %d\n",i,global_sim_i.len);
}
//getchar();
inter_save_a(&global_sim_i,"./","ittr_sim_i.dat");
inter_save_a(&global_in,"./","ittr_in.dat");
inter_save_a(&global_dos,"./","ittr_dos.dat");
inter_save_a(&global_error,"./","ittr_error.dat");
//getchar();

printf("exit expb_df\n");
return GSL_SUCCESS;
}



int pt_expb_f (const gsl_vector * x, void *data, gsl_vector * f)
{
printf("expb_f\n");
double *sigma = global_sigma.data;
pt_update_dos(x);
get_pt_error();

size_t i;

for (i = 0; i < global_error.len; i++)
{
	/* Model Yi = A * exp(-lambda * i) + b */
	//double Yi = main_fit.data[i];//A * exp (-lambda * t) + b;

		gsl_vector_set (f, i, (global_error.data[i])/sigma[i]);

	//printf("test=%le\n",(global_error.data[i]));


}
printf("exit expb_f\n\n");
return GSL_SUCCESS;
}


int pt_expb_fdf (const gsl_vector * x, void *data, gsl_vector * f, gsl_matrix * J)
{
printf("expb_fdf\n");
pt_expb_f (x, data, f);
pt_expb_df (x, data, J);
printf("exit expb_fdf\n");
//getchar();

return GSL_SUCCESS;
}



void pt_extract(char *outputpath,char *infile,double T,int pt_fit,double start,double stop,double mull,double dosstart,double dosstop,int maxittr,double chop,double chop1,double norm_point,double norm_E_pos,double sigmain,double v0,double norm_E_mull)
{
static struct istruct dos_out;
static struct istruct global_dos_pt;
int i;
double sigma=1e-16;
double t0=7.2e-5;
double Q0=0.0;
double t=0.0;
double E=0.0;
double L=1e-6;
double V=60;
double F=V/L;
double g0=0.0;
double mu=L/(t0*F);

inter_load_a(&global_in,infile,0,1);
inter_chop(&global_in,1e-10, 100);
inter_save_seg(&global_in,outputpath,"outtrans",9);
Q0=inter_intergrate(&global_in);
double div=inter_get(&global_in,norm_point);
g0=(2.0*norm_point*div)/(Q0*t0*v0);
printf("%le\n",g0);
//getchar();

inter_rescale(&global_in,1.0, 1.0/div);
inter_save_a(&global_in,outputpath,"outtrans.dat");
inter_new(&global_dos,global_in.len);
inter_new(&global_dos_pt,global_in.len);

for (i=0;i<global_in.len;i++)
{
t=global_in.x[i];
E=kb*T*log(v0*t)/q;
//printf("%le %le\n",v0,E);
//getchar();
//printf("%le %le\n",E,t);
//getchar();
global_dos.x[i]=E;
global_dos.data[i]=(Q0*t0*v0)*(global_in.data[i]*t);
global_dos_pt.x[i]=E;
global_dos_pt.data[i]=(Q0*t0*v0)/(global_in.data[i]*t);
}

if (norm_E_pos!=-1.0)
{
div=inter_get(&global_dos,norm_E_pos);
inter_rescale(&global_dos,1.0, norm_E_mull/div);
div=inter_get(&global_dos_pt,norm_E_pos);
inter_rescale(&global_dos_pt,1.0, norm_E_mull/div);
}

inter_save_seg(&global_dos,outputpath,"outdos",9);
inter_save_a(&global_dos,outputpath,"outdos.dat");

inter_save_a(&global_dos_pt,outputpath,"outdos_pt.dat");
V=60;
F=V/L;
t0=L/(mu*F);
for (i=0;i<global_in.len;i++)
{
E=global_dos.x[i]*q;
t=(1.0/v0)*exp(E/(kb*T));//log(v0*t)/q;
//getchar();
global_in.x[i]=t;
global_in.data[i]=global_dos.data[i]*(Q0*t0*v0)/t;
}

inter_save_a(&global_in,outputpath,"regen.dat");
inter_free(&global_in);
inter_free(&global_dos);
inter_free(&global_dos_pt);

if (pt_fit==TRUE)
{
struct istruct in;
struct istruct in1;
inter_load_a(&in1,infile,0,1);
//inter_mod(&in1);

if (dolog==TRUE) inter_log_y(&in1);
inter_save_a(&in1,outputpath,"save_new_mesh1.dat");

int dos_len=mypoints;
int tpoints=mypoints;
inter_new(&in,2000);
inter_to_log_mesh(&in,&in1);
//inter_to_new_mesh(&in1,&in);
inter_save_a(&in,outputpath,"save_new_mesh2.dat");
inter_free(&in1);
double E_start=dosstart;
double E_stop=dosstop;
//printf("%le %le\n",E_start,E_stop);
//getchar();
double E_pos=0.0;
double E_delta=0.0;

inter_new(&global_r,dos_len);
inter_new(&global_A,dos_len);
inter_new(&global_dos,dos_len);



E_pos=E_start;
E_delta=(E_stop-E_start)/((double)dos_len);
for (i=0;i<global_r.len;i++)
{

	global_r.x[i]=E_pos;
	global_A.x[i]=E_pos;
	global_dos.x[i]=E_pos;
	global_r.data[i]=v0*exp((-E_pos*q)/(T*kb));
	global_dos.data[i]=mull;//*exp((E_pos)/(19e-3));
	E_pos+=E_delta;
}


inter_new(&global_in,tpoints);


inter_smooth(&in,4);
inter_chop(&in,start, stop);

inter_to_log_mesh(&global_in,&in);


inter_save_a(&global_in,outputpath,"pt_exp_i.dat");
inter_free(&in);

inter_new(&global_sim_i,global_in.len);
inter_new(&global_error,global_in.len);
inter_new(&global_sigma,global_in.len);
get_pt_error();


const gsl_multifit_fdfsolver_type *Tsolve;
gsl_multifit_fdfsolver *s;
int status;
unsigned iter = 0;
const size_t n = global_sim_i.len;
const size_t p = global_dos.len;
//printf("%d %d\n",n,p);
//getchar();

/*pt_Ti =    malloc(pt_N*sizeof(int));
pt_Tj =    malloc(pt_N*sizeof(int));
pt_Tx = malloc(pt_N*sizeof(double));
pt_b = malloc(pt_M*sizeof(double));
do
{
do_newton();
getchar();
}while(1);
exit(0);*/
gsl_matrix *covar = gsl_matrix_alloc (n, p);

for (i=0;i<global_sigma.len;i++)
{
	global_sigma.data[i] = sigmain;
}

struct pt_data d = { n,  global_in.data, global_sigma.data};

gsl_multifit_function_fdf f;



gsl_vector_view x = gsl_vector_view_array (((global_dos.data)), p);

//printf("test=%le %le\n",gsl_vector_get(&(x.vector), 1),main_g.data[fit_start+1]);
//getchar();

f.f = &pt_expb_f;
f.df = &pt_expb_df;
f.fdf = &pt_expb_fdf;
f.n = n;
f.p = p;
f.params = &d;


Tsolve = gsl_multifit_fdfsolver_lmsder;
s = gsl_multifit_fdfsolver_alloc (Tsolve, n, p);
//printf("%ld %ld\n",n,p);
//getchar();
gsl_multifit_fdfsolver_set (s, &f, &x.vector);

do
{
   iter++;
   status = gsl_multifit_fdfsolver_iterate (s);

   printf ("status = %s\n", gsl_strerror (status));

           if (status)
             break;

   status = gsl_multifit_test_delta (s->dx, s->x,1e-10, 1e-10);



//inter_smooth(&global_dos,5);
//inter_smooth(&global_dos,5);

inter_copy(&dos_out,&global_dos,TRUE);
div=inter_get(&dos_out,0.27);
inter_rescale(&dos_out,1.0, 1.0/div);
inter_chop(&dos_out,chop1, chop);
inter_save_a(&dos_out,outputpath,"pt_fit_dos.dat");

inter_free(&dos_out);

inter_save_a(&global_sim_i,outputpath,"pt_fit_i.dat");

printf("~log = %d ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n",iter);

}
while (status == GSL_CONTINUE && iter < maxittr);

gsl_multifit_covar (s->J, 0.0, covar);

#define FIT(i) gsl_vector_get(s->x, i)
#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

printf ("status = %s\n", gsl_strerror (status));

gsl_multifit_fdfsolver_free (s);
gsl_matrix_free (covar);


inter_save_a(&global_in,outputpath,"g_out.dat");
inter_save_a(&global_sim_i,outputpath,"sim_i.dat");


inter_free(&global_error);
inter_free(&global_sim_i);
inter_free(&global_A);
inter_free(&global_r);
inter_free(&global_in);
inter_free(&global_dos);
inter_free(&global_sigma);

}

}



